﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.BinPacking.Interfaces;

namespace HeuristicLab.Encodings.PackingEncoding.PackingSequence {
  [Item("SequenceMoveSoftTabuCriterion", "<Missing>")]
  [StorableClass]
  public class SequenceMoveSoftTabuCriterion : SingleSuccessorOperator, IPackingSequenceMoveOperator, ITabuChecker {
    public override bool CanChangeName {
      get { return false; }
    }
    public ILookupParameter<IPackingMove> PackingMoveParameter {
      get { return (ILookupParameter<IPackingMove>)Parameters["PackingMove"]; }
    }
    public ILookupParameter<PackingSequenceEncoding> PackingSequenceParameter {
      get { return (ILookupParameter<PackingSequenceEncoding>)Parameters["PackingSequence"]; }
    }
    public ILookupParameter<ItemList<IItem>> TabuListParameter {
      get { return (ILookupParameter<ItemList<IItem>>)Parameters["TabuList"]; }
    }
    public ILookupParameter<BoolValue> MoveTabuParameter {
      get { return (ILookupParameter<BoolValue>)Parameters["MoveTabu"]; }
    }
    public IValueLookupParameter<BoolValue> MaximizationParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["Maximization"]; }
    }
    public ILookupParameter<DoubleValue> MoveQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["MoveQuality"]; }
    }
    public ValueParameter<BoolValue> UseAspirationCriterionParameter {
      get { return (ValueParameter<BoolValue>)Parameters["UseAspirationCriterion"]; }
    }

    [StorableConstructor]
    protected SequenceMoveSoftTabuCriterion(bool deserializing) : base(deserializing) { }
    protected SequenceMoveSoftTabuCriterion(SequenceMoveSoftTabuCriterion original, Cloner cloner) : base(original, cloner) { }
    public SequenceMoveSoftTabuCriterion()
      : base() {
      Parameters.Add(new LookupParameter<IPackingMove>("PackingMove", "The move to evaluate."));
      Parameters.Add(new LookupParameter<PackingSequenceEncoding>("PackingSequence", "The solution to evaluate."));
      Parameters.Add(new LookupParameter<BoolValue>("MoveTabu", "The variable to store if a move was tabu."));
      Parameters.Add(new LookupParameter<ItemList<IItem>>("TabuList", "The tabu list."));
      Parameters.Add(new ValueLookupParameter<BoolValue>("Maximization", "True if the problem is a maximization problem, else if it is a minimization problem."));
      Parameters.Add(new LookupParameter<DoubleValue>("MoveQuality", "The quality of the current move."));
      Parameters.Add(new ValueParameter<BoolValue>("UseAspirationCriterion", "Whether to use the aspiration criterion or not.", new BoolValue(true)));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SequenceMoveSoftTabuCriterion(this, cloner);
    }

    public override IOperation Apply() {
      ItemList<IItem> tabuList = TabuListParameter.ActualValue;
      SequenceMove move = PackingMoveParameter.ActualValue as SequenceMove;
      PackingSequenceEncoding PackingSequence = PackingSequenceParameter.ActualValue;
      int length = PackingSequence.PackingSequence.Length;
      double moveQuality = MoveQualityParameter.ActualValue.Value;
      bool maximization = MaximizationParameter.ActualValue.Value;   
      bool useAspiration = UseAspirationCriterionParameter.Value.Value;
      bool isTabu = IsMoveTabu(tabuList, move, PackingSequence, moveQuality, maximization, useAspiration);

      MoveTabuParameter.ActualValue = new BoolValue(isTabu);
      return base.Apply();
    }

    private bool IsMoveTabu(ItemList<IItem> tabuList, SequenceMove move, PackingSequenceEncoding PackingSequence, double moveQuality, bool maximization, bool useAspiration) {
      bool isTabu = false;
      var attributes = tabuList.FindAll(ta => ta.GetType().Equals(move.GetMoveAttributeType()));

      foreach (IItem attr in attributes) {
        SequenceMoveAttribute gvAttr = attr as SequenceMoveAttribute;
        if (gvAttr != null
          && (!useAspiration
              || maximization && moveQuality <= gvAttr.MoveQuality
              || !maximization && moveQuality >= gvAttr.MoveQuality)) {

          isTabu = move.BelongsToAttribute(gvAttr, false);
        }
        if (isTabu) break;
      }

      return isTabu;
    }
  }
}
